from __future__ import annotations

import argparse
import os
import re
import sys
from collections.abc import Iterable
from difflib import unified_diff
from io import StringIO
from tempfile import TemporaryDirectory
from typing import IO

from inifix.io import _split_tokens, load

PADDING_SIZE = 2


def _format_section(data: str) -> str:
    lines = data.splitlines()
    contents = []
    comments = []
    parameters = []
    values = []
    for line in lines:
        content, _, comment = line.partition("#")
        content = content.strip()
        if not content and not comment:
            continue
        contents.append(content)
        comments.append(comment.strip())
        if content.startswith("[") or not content:
            continue
        parameter, *value = _split_tokens(content)
        parameters.append(parameter)
        values.append(value)

    column_sizes: list[int] = []
    if not parameters:
        max_name_size = 0
    else:
        max_name_size = max(len(parameter) for parameter in parameters)
        num_value_cols = max(len(_) for _ in values)

        while len(column_sizes) < num_value_cols:
            new_column_idx = len(column_sizes)
            size = 0
            for v in values:
                if len(v) < new_column_idx + 1:
                    continue
                size = max(size, len(v[new_column_idx]))
            column_sizes.append(size + PADDING_SIZE)

    padded_name_col_size = max_name_size + 2 * PADDING_SIZE
    content_size = padded_name_col_size + sum(column_sizes) + PADDING_SIZE

    new_lines = []
    parameter_idx = 0
    for content, comment in zip(contents, comments):
        new_line = ""
        offset = 0
        if content:
            if content.startswith("["):
                new_line += f"{content}\n"
            else:
                new_line += parameters[parameter_idx].ljust(padded_name_col_size)
                for val, size in zip(values[parameter_idx], column_sizes):
                    new_line += val.ljust(size)
                parameter_idx += 1
                new_line = new_line.strip()
                offset = content_size - len(new_line)
        if comment:
            comm = f"# {comment}"
            new_line += comm.rjust(offset + len(comm))
        new_lines.append(new_line)
    res = "\n".join(new_lines)

    return res


def _finalize(res: str) -> str:
    # compress any duplicate new lines
    res = re.sub("\n+", "\n", res)
    # add one empty line before a new section
    res = re.sub(r"(.?)\n\[", r"\1\n\n[", res)
    # ensure there's exactly one newline at the EOF
    res = res.rstrip("\n")
    res += "\n"
    return res


def _iter_sections(fh: IO[str]) -> Iterable[str]:
    line = fh.readline()
    while line != "":
        content = [line]
        while (line := fh.readline()) != "" and not line.startswith("["):
            content.append(line)
        yield "".join(content)


def iniformat(fh: IO[str] | str, /) -> str:
    if isinstance(fh, str):
        fh = StringIO(fh)
    content = []
    for s in _iter_sections(fh):
        content.append(_format_section(s))
    return _finalize("\n".join(content))


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs="+")
    parser.add_argument(
        "--diff",
        action="store_true",
        help="Print the unified diff to stdout instead of editing files inplace",
    )
    parser.add_argument(
        "--report-noop",
        action="store_true",
        help="Explicitly log noops for files that are already formatted",
    )
    parser.add_argument(
        "--skip-validation",
        dest="validate",
        action="store_false",
        help="Skip validation step (formatting unvalidated data may lead to undefined behaviour)",
    )

    args = parser.parse_args(argv)

    retv = 0

    for file in args.files:
        if not os.path.isfile(file):
            print(f"Error: could not find {file}", file=sys.stderr)
            retv = 1
            continue

        if args.validate:
            try:
                load(file)
            except ValueError as exc:
                print(f"Error: {exc}", file=sys.stderr)
                retv = 1
                continue

        with open(file, mode="rb") as fh:
            data = fh.read().decode("utf-8")
            # make sure newlines are always decoded as \n, even on windows
            data = data.replace("\r\n", "\n")

        fmted_data = iniformat(data)

        if fmted_data == data:
            if args.report_noop:
                # printing to stderr so that we can pipe into cdiff in --diff mode
                print(f"{file} is already formatted", file=sys.stderr)
            continue
        else:
            retv = 1

        if args.diff:
            for line in unified_diff(
                data.splitlines(), fmted_data.splitlines(), fromfile=file
            ):
                line = line.removesuffix("\n")
                print(line)
            print("\n")
        else:
            print(f"Fixing {file}", file=sys.stderr)
            if not os.access(file, os.W_OK):
                print(
                    f"Error: could not write to {file} (permission denied)",
                    file=sys.stderr,
                )
                retv = 1
                continue

            with TemporaryDirectory(dir=os.path.dirname(file)) as tmpdir:
                tmpfile = os.path.join(tmpdir, "ini")
                with open(tmpfile, "wb") as bfh:
                    bfh.write(fmted_data.encode("utf-8"))
                # this may still raise an error in the unlikely case of a race condition
                # (if permissions are changed between the look and the leap), but we
                # won't try to catch it unless it happens in production, because it is
                # difficult to test systematically.
                os.replace(tmpfile, file)

    return retv


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



from __future__ import annotations

import os
import re
from collections.abc import Mapping
from functools import partial
from io import BufferedIOBase, IOBase
from typing import Any, Callable, cast

from inifix._more import always_iterable
from inifix._typing import InifixConfT, IterableOrSingle, PathLike, Scalar, StrLike
from inifix.enotation import ENotationIO
from inifix.validation import SCALAR_TYPES, validate_inifile_schema

__all__ = ["load", "loads", "dump", "dumps"]

SECTION_REGEXP = re.compile(r"\[(?P<title>[^(){}\[\]]+)\]\s*")


def _is_numeric(s: str) -> bool:
    try:
        float(s)
    except ValueError:
        return False
    else:
        return True


class Section(dict):
    def __init__(
        self,
        data: Mapping[str, IterableOrSingle[Scalar]] | None = None,
        /,
        *,
        name: str | None = None,
    ) -> None:
        super().__init__()
        if data is not None:
            for k, v in data.items():
                self[k] = v
        self.name = name

    def __setitem__(self, k: Any, v: Any) -> None:
        if not isinstance(k, str):
            raise TypeError(f"Expected str keys. Received invalid key: {k}")

        if not (
            isinstance(v, SCALAR_TYPES)
            or (isinstance(v, list) and all(isinstance(_, SCALAR_TYPES) for _ in v))
        ):
            raise TypeError(
                "Expected all values to be scalars or lists of scalars. "
                f"Received invalid values {v}"
            )

        super().__setitem__(k, v)

    def _dump_to(self, storage: InifixConfT) -> None:
        if self.name is None:
            storage.update(self)
        else:
            storage[self.name] = dict(self)


# load helper functions


# this is more efficient than running str.partition in a loop
_SPLIT_COMMENTS = partial(str.split, sep="#", maxsplit=1)


def _normalize_data(data: StrLike) -> list[str]:
    # normalize text body `data` to parsable text lines
    if isinstance(data, bytes):
        data = data.decode("utf-8")
    return [line.strip() for (line, *_) in map(_SPLIT_COMMENTS, data.splitlines())]


_TOKEN = re.compile(r"""'[^']*'|"[^"]*"|\S+""")


def _split_tokens(data: str) -> list[str]:
    return _TOKEN.findall(data)


_RE_CASTERS: list[tuple[re.Pattern, Callable[[str], Any]]] = [
    (re.compile(r"(true|yes)", re.I), lambda _: True),
    (re.compile(r"(false|no)", re.I), lambda _: False),
    (re.compile(r"^'.*'$"), lambda s: s[1:-1]),
    (re.compile(r'^".*"$'), lambda s: s[1:-1]),
]


def _auto_cast(s: str) -> Any:
    try:
        f = float(s)
    except ValueError:
        pass
    else:
        if f.is_integer():
            return int(f)
        else:
            return f

    for regexp, caster in _RE_CASTERS:
        if regexp.fullmatch(s):
            return caster(s)

    return s


def _tokenize_line(
    line: str, line_number: int, filename: str | None
) -> tuple[str, list[Scalar]]:
    key, *raw_values = _split_tokens(line)
    if not raw_values:
        if filename is None:
            raise ValueError(f"Failed to parse line {line_number}: {line!r}")
        else:
            raise ValueError(f"Failed to parse {filename}:{line_number}:\n{line}")

    return key, [_auto_cast(v) for v in raw_values]


def _from_string(
    data: StrLike, *, parse_scalars_as_lists: bool, filename: str | None = None
) -> InifixConfT:
    # see https://github.com/python/mypy/issues/6463
    container: InifixConfT = {}  # type: ignore[assignment]
    lines = _normalize_data(data)
    section = Section()  # the default target is a nameless section
    for line_number, line in enumerate(lines, start=1):
        if not line:
            continue
        match = SECTION_REGEXP.fullmatch(line)
        if match is not None:
            if section:
                section._dump_to(container)
            section = Section(name=match["title"])
            continue

        values: Scalar | list[Scalar]
        key, values = _tokenize_line(line, filename=filename, line_number=line_number)
        if (not parse_scalars_as_lists) and len(values) == 1:
            values = values[0]
        section[key] = values
    section._dump_to(container)
    return container


def _from_file_descriptor(file: IOBase, *, parse_scalars_as_lists: bool) -> InifixConfT:
    filename = str(getattr(file, "name", repr(file)))
    data = file.read()
    lines = _normalize_data(data)
    if not "".join(lines):
        raise ValueError(f"{filename!r} appears to be empty.")
    return _from_string(
        data, filename=filename, parse_scalars_as_lists=parse_scalars_as_lists
    )


def _from_path(file: PathLike, *, parse_scalars_as_lists: bool) -> InifixConfT:
    file = os.fspath(file)
    with open(file, "rb") as fh:
        return _from_file_descriptor(fh, parse_scalars_as_lists=parse_scalars_as_lists)


# dump helper functions


def _encode(v: Scalar) -> str:
    if isinstance(v, float):
        return ENotationIO.encode_preferential(v)
    elif isinstance(v, str) and (
        re.search(r"\s", v) is not None
        or v in ("true", "t", "false", "f")
        or _is_numeric(v)
    ):
        return repr(v)
    else:
        return str(v)


def _write(content: str, buffer: IOBase) -> None:
    if isinstance(buffer, BufferedIOBase):
        buffer.write(content.encode("utf-8"))
    else:
        buffer.write(content)


def _write_line(key: str, values: IterableOrSingle[Scalar], buffer: IOBase) -> None:
    val_repr = [_encode(v) for v in always_iterable(values)]
    _write(f"{key} {'  '.join(list(val_repr))}\n", buffer)


def _write_to_buffer(data: InifixConfT, buffer: IOBase) -> None:
    last = len(data) - 1
    for i, (key, val) in enumerate(data.items()):
        if not isinstance(val, dict):
            _write_line(key, val, buffer)
            continue
        _write(f"[{key}]\n", buffer)
        for k, v in val.items():
            _write_line(k, v, buffer)
        if i < last:
            _write("\n", buffer)


def _write_to_file(data: InifixConfT, file: PathLike, /) -> None:
    if os.path.exists(file) and not os.access(file, os.W_OK):
        raise PermissionError(f"Cannot write to {file} (permission denied)")

    from tempfile import TemporaryDirectory

    with TemporaryDirectory(dir=os.path.dirname(file)) as tmpdir:
        tmpfile = os.path.join(tmpdir, "ini")
        with open(tmpfile, "wb") as fh:
            _write_to_buffer(data, fh)
        os.replace(tmpfile, file)


def load(
    source: InifixConfT | PathLike | IOBase,
    /,
    *,
    parse_scalars_as_lists: bool = False,
    skip_validation: bool = False,
) -> InifixConfT:
    """
    Parse data from a file, or a dict.

    Parameters
    ----------
    source: any of the following
        - a dict (has to be inifix format-compliant)
        - the name of a file to read from, (str, bytes or os.PathLike)
        - a readable handle. Both text and binary file modes are supported,
          though binary is prefered.
          In binary mode, we assume UTF-8 encoding.

    parse_scalars_as_lists: bool
        if set to True, all values will be parses as lists of scalars,
        even for parameters comprised of a single scalar.

    skip_validation: bool
        if set to True, input is not validated. This can be used to speedup io operations
        trusted input or to work around bugs with the validation routine. Use with caution.

    See Also
    --------
    inifix.loads
    """
    if isinstance(source, IOBase):
        source = _from_file_descriptor(
            source, parse_scalars_as_lists=parse_scalars_as_lists
        )
    elif isinstance(source, (str, bytes, os.PathLike)):
        source = _from_path(source, parse_scalars_as_lists=parse_scalars_as_lists)

    source = cast(Mapping, source)

    if not skip_validation:
        validate_inifile_schema(source)
    source = cast(InifixConfT, source)
    return source


def loads(
    source: str,
    /,
    *,
    parse_scalars_as_lists: bool = False,
    skip_validation: bool = False,
) -> InifixConfT:
    """
    Parse data from a string.

    Parameters
    ----------
    source: str
        The content of a parameter file (has to be inifix format-compliant)

    parse_scalars_as_lists: bool
        if set to True, all values will be parses as lists of scalars,
        even for parameters comprised of a single scalar.

    skip_validation: bool
        if set to True, input is not validated. This can be used to speedup io operations
        trusted input or to work around bugs with the validation routine. Use with caution.

    See Also
    --------
    inifix.load
    """
    retv = _from_string(source, parse_scalars_as_lists=parse_scalars_as_lists)

    if not skip_validation:
        validate_inifile_schema(retv)
    return retv


def dump(
    data: InifixConfT, /, file: PathLike | IOBase, *, skip_validation: bool = False
) -> None:
    """
    Write data to a file.

    Parameters
    ----------
    data: dict
        has to be inifix format-compliant

    file: any of the following
        - the name of a file to write to (str, bytes or os.PathLike)
        - a writable handle. Both text and binary file modes are supported,
          though binary is prefered.
          In binary mode, data is encoded as UTF-8.

    skip_validation: bool
        if set to True, input is not validated. This can be used to speedup io operations
        trusted input or to work around bugs with the validation routine. Use with caution.

    See Also
    --------
    inifix.dumps
    """
    if not skip_validation:
        validate_inifile_schema(data)

    if isinstance(file, IOBase):
        _write_to_buffer(data, file)
    else:
        _write_to_file(data, file)


def dumps(data: InifixConfT, /, *, skip_validation: bool = False) -> str:
    """
    Convert data to a string.

    Parameters
    ----------
    data: dict
        has to be inifix format-compliant

    skip_validation: bool
        if set to True, input is not validated. This can be used to speedup io operations
        trusted input or to work around bugs with the validation routine. Use with caution.

    Returns
    -------
    string representing the content of a parameter file

    See Also
    --------
    inifix.dump
    """
    from io import BytesIO

    s = BytesIO()
    dump(data, file=s, skip_validation=skip_validation)
    return s.getvalue().decode("utf-8")



import pytest

import inifix
from inifix.io import _auto_cast

BASE_BOOLS = [
    ("True", True),
    ("true", True),
    ("Yes", True),
    ("yes", True),
    ("False", False),
    ("false", False),
    ("No", False),
    ("no", False),
]


@pytest.mark.parametrize("s, expected", BASE_BOOLS)
def test_bool_cast(s, expected):
    assert _auto_cast(s) is expected


@pytest.mark.parametrize("s", ["tdsk", "1213", "Treu", "Flsae", "flkj"])
def test_bool_cast_invalid(s):
    assert type(_auto_cast(s)) is not bool


@pytest.mark.parametrize("s, expected", BASE_BOOLS)
def test_bool_cast_integration(s, expected, tmp_path):
    with open(tmp_path / "dummy.ini", "w") as fh:
        fh.write(f"dummy {s}")
    d = inifix.load(tmp_path / "dummy.ini")
    assert d == {"dummy": expected}


@pytest.mark.parametrize(
    "s, expected",
    [
        ("1.", 1),
        ("1.0", 1),
        ("1e3", 1000),
        ("+1.", 1),
        ("+1.0", 1),
        ("+1e3", 1000),
        ("-1.", -1),
        ("-1.0", -1),
        ("-1e3", -1000),
    ],
)
def test_float_cast(s, expected):
    res = _auto_cast(s)
    assert type(res) is type(expected)
    assert res == expected



import pytest
from more_itertools import unzip

from inifix.io import dump, load
from inifix.validation import validate_inifile_schema

INVADLID_SCHEMAS, INVALID_SCHEMAS_IDS = unzip(
    (
        ({1: {"param": 1}}, "section_no_str"),
        ({"section": {"": 1}}, "empty_param_name"),
        ({"section": {"§": 1}}, "invalid_param_name-1"),
        ({"section": {"a§": 1}}, "invalid_param_name-2"),
        ({"section": {1: "one"}}, "param_name_no_str"),
        ({"section": [[1, 2], [3, 4]]}, "non_flat_section"),
        ({"section": {"param": [[1, 2], [3, 4]]}}, "nested_non_flat_sequence"),
        ({"section": {"subsection": {"subsubsection": 1}}}, "too_deep_nesting"),
    )
)


@pytest.fixture(params=INVADLID_SCHEMAS, ids=INVALID_SCHEMAS_IDS)
def invalid_conf(request):
    return request.param


def test_validate_known_files(inifile):
    conf = load(inifile)
    validate_inifile_schema(conf)


def test_dump_invalid_conf(invalid_conf, tmp_path):
    with pytest.raises(ValueError, match=r"^(Invalid schema)"):
        dump(invalid_conf, tmp_path / "save.ini")



import os
from collections.abc import Iterable
from typing import AnyStr, TypeVar, Union

# not quite typing.AnyStr : this is not a constrained type variable
StrLike = Union[str, bytes]

PathLike = Union[AnyStr, os.PathLike[AnyStr]]

T = TypeVar("T")
Scalar = Union[int, float, bool, str]
IterableOrSingle = Union[Iterable[T], T]

# these types are used to validate schemas internally at type-checking time
SectionT = dict[str, IterableOrSingle[Scalar]]
InifixConfT = Union[dict[str, SectionT], SectionT]
